/****************************************************************************/
// camdata.c
//
//  - populate & refresh container with photo data
//  - save & delete files using container data
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#include "camapp.h"
#include "camusb.h"

/****************************************************************************/

typedef struct _CAMSaveFile {
    CAMRecPtr       pcr;
    BOOL            fHalt;
    HWND            hCnr;
    HWND            hBtn1;
    HWND            hBtn2;
    HWND            hBtn3;
    HWND            hBtn4;
    HWND            hEdit;
    HWND            hChk;
    ULONG           cntSave;
    ULONG           cbSave;
    ULONG           cntErase;
    ULONG           cbErase;
    ULONG           ulMsg;
    ULONG           flags;
    char            szPath[524];
} CAMSaveFile;

typedef CAMSaveFile *CAMSaveFilePtr;

#define CAMSAVE_REPONE      0x1
#define CAMSAVE_REPALL      0x2
#define CAMSAVE_REPMASK     0x3
#define CAMSAVE_RETRY       0x10
#define CAMSAVE_EDIT        0x20
#define CAMSAVE_ERRMASK     0x30
#define CAMSAVE_SAVE        0x100
#define CAMSAVE_SAVEERASE   0x200
#define CAMSAVE_SAVEMARK    0x400
#define CAMSAVE_ERASE       0x800

#define CAMSAVE_RC_INIT        ((ULONG)-1)
#define CAMSAVE_RC_OK          0
#define CAMSAVE_RC_TOOLONG     1
#define CAMSAVE_RC_FILEEXISTS  2
#define CAMSAVE_RC_BADPATH     3
#define CAMSAVE_RC_GETDATA     4
#define CAMSAVE_RC_DOSOPEN     5
#define CAMSAVE_RC_DOSWRITE    6
#define CAMSAVE_RC_DELDATA     7
#define CAMSAVE_RC_BADCREATE   8
#define CAMSAVE_RC_CANCELLED   9
#define CAMSAVE_RC_DONE        10


/****************************************************************************/

// this struct combines an FEA2LIST, an FEA2, plus additional fields
// needed to write a .SUBJECT EA in the correct EAT_ASCII format;

#pragma pack(1)
    typedef struct _CAMSubjectEA {
        struct {
            ULONG   ulcbList;
            ULONG   uloNextEntryOffset;
            BYTE    bfEA;
            BYTE    bcbName;
            USHORT  uscbValue;
            char    chszName[sizeof(".SUBJECT")];
        } hdr;
        struct {
            USHORT  usEAType;
            USHORT  usDataLth;
        } info;
        char    data[32];
    } CAMSubjectEA;
#pragma pack()

/****************************************************************************/

//BOOL            SaveChanges( HWND hwnd, ULONG ulMsg);
MRESULT _System SaveDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
LONG            InitSaveDlg( HWND hwnd, MPARAM mp2);
BOOL            SaveFiles( HWND hwnd);
MRESULT _System SaveFilesDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
LONG            InitSaveFilesDlg( HWND hwnd, MPARAM mp2);
void            StartSavingFiles( HWND hwnd, ULONG flags);
void            SaveNextFile( HWND hwnd, CAMSaveFilePtr pCSF);
void            DoneSavingFiles( HWND hwnd, CAMSaveFilePtr pCSF);
void            EraseFile( HWND hwnd, CAMSaveFilePtr pCSF);
ULONG           EraseFileInternal( HWND hwnd, CAMSaveFilePtr pCSF);
void            SaveFile( HWND hwnd, CAMSaveFilePtr pCSF);
void            SetSaveControls( HWND hwnd, CAMSaveFilePtr pCSF, ULONG ulErr);
ULONG           BuildFilename( CAMRecPtr pcr, char * pszName);
void            BuildSubjectEA( CAMRecPtr pcr, CAMSubjectEA * pEA);
void            CreateSavePath( HWND hwnd, CAMSaveFilePtr pCSF);
ULONG           NextMarkedRec( HWND hCnr, CAMRecPtr * pprcIn, ULONG flags);

//ULONG           GetData( HWND hwnd, BOOL fSync);
void            AdjustList( CAMObjectHandlePtr pHandles);
ULONG           GetList( HWND hCnr, CAMObjectHandlePtr pHandles);
ULONG       	SyncList( HWND hCnr, CAMObjectHandlePtr pHandles);
void            SetData( CAMRecPtr pcr, ULONG ulHndl, CAMObjectInfoPtr pObjInfo);
void            NextRec( HWND hCnr, CAMRecPtr * ppcrIn);

/****************************************************************************/

typedef union _LRGB {
    LONG    l;
    RGB     rgb;
} LRGB;

static LRGB     lrgbEditBkgd        = {0x00ffffff};
static LRGB     lrgbWhite           = {0x00ffffff};

static BOOL     fInitSaveFilesMsgs  = FALSE;
static BOOL     fInitColMsgs        = FALSE;
static char     szX[]               = "xXx";

static char *   pszMSG_Erasing      = szX;
static char *   pszMSG_Saving       = szX;
static char *   pszMSG_SavedErased  = szX;
static char *   pszBTN_Retry        = szX;
static char *   pszBTN_Skip         = szX;
static char *   pszBTN_Stop         = szX;
static char *   pszBTN_Close        = szX;
static char *   pszBTN_Replace      = szX;
static char *   pszBTN_Create       = szX;

static char *   pszCOL_TooBig       = szX;
static char *   aszDay[]            = {szX, szX, szX, szX, szX, szX, szX};

static CAMNLSPSZ nlsColMsgs[] = {
    { &pszCOL_TooBig,       COL_TooBig },
    { &aszDay[0],           COL_Sun },
    { &aszDay[1],           COL_Mon },
    { &aszDay[2],           COL_Tue },
    { &aszDay[3],           COL_Wed },
    { &aszDay[4],           COL_Thu },
    { &aszDay[5],           COL_Fri },
    { &aszDay[6],           COL_Sat },
    { 0,                    0 }};

static CAMNLSPSZ nlsSaveFilesMsgs[] = {
    { &pszMSG_Erasing,      MSG_Erasing },
    { &pszMSG_Saving,       MSG_Saving },
    { &pszMSG_SavedErased,  MSG_SavedErased },
    { &pszBTN_Retry,        BTN_Retry },
    { &pszBTN_Skip,         BTN_Skip },
    { &pszBTN_Stop,         BTN_Stop },
    { &pszBTN_Close,        BTN_Close },
    { &pszBTN_Replace,      BTN_Replace },
    { &pszBTN_Create,       BTN_Create },
    { 0,                    0 }};

static CAMNLS   nlsSaveDlg[] = {
    { IDC_SAVECHGS,         DLG_SaveChanges },
    { IDC_SAVEDONT,         DLG_DontErase },
    { IDC_SAVELATER,        DLG_EraseLater },
    { IDC_SAVEAFTER,        DLG_EraseAfterSave },
    { IDC_YES,              BTN_Yes },
    { IDC_NO,               BTN_No },
    { IDC_CANCEL,           BTN_Cancel },
    { 0,                    0 }};

static CAMNLS   nlsSaveFilesDlg[] = {
    { IDC_REPLACEALL,       DLG_ReplaceAll },
    { 0,                    0 }};

/****************************************************************************/
/****************************************************************************/

BOOL        SaveChanges( HWND hwnd, ULONG ulMsg)

{
    STATSTUFF *     pss;

    pss = GetStatStuff();

    // if nothing to do, exit
    if (pss->ulSavCnt == 0 && pss->ulDelCnt == 0)
        return (TRUE);

    WinSendMsg( hwnd, CAMMSG_DLGSTATUS, (MP)IDD_SAVE, (MP)TRUE);

    // show confirmation dialog
    return (!WinLoadDlg(
                HWND_DESKTOP,       // parent-window
                hwnd,               // owner-window
                SaveDlgProc,        // dialog proc
                NULLHANDLE,         // EXE module handle
                (ulMsg ? IDD_SAVEEXIT : IDD_SAVE),   // dialog id
                (PVOID)ulMsg));     // pointer to create params
}

/****************************************************************************/

MRESULT _System SaveDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{
    switch (msg)
    {
        case WM_COMMAND:
            switch (SHORT1FROMMP( mp1)) {

                case IDC_YES:
                    if (SaveFiles( hwnd))
                        // use mp1 to tell WM_CLOSE handler
                        // not to reenable the owner window
                        WinSendMsg( hwnd, WM_CLOSE, (MP)TRUE, 0);
                    else
                        WinSendMsg( hwnd, WM_COMMAND, (MP)IDC_NO, 0);
                    break;

                case IDC_NO: {
                    ULONG   ulMsg = WinQueryWindowULong( hwnd, QWL_USER);

                    if (ulMsg) {
                        STATSTUFF * pss = GetStatStuff();
                        pss->ulSavCnt = 0;
                        pss->ulSavSize = 0;
                        pss->ulDelCnt = 0;
                        pss->ulDelSize = 0;
                        WinPostMsg( WinQueryWindow( hwnd, QW_OWNER),
                                    LOUSHORT(ulMsg),
                                    MPFROM2SHORT( HIUSHORT(ulMsg), 0), 0);
                    }

                    WinSendMsg( hwnd, WM_CLOSE, 0, 0);
                    break;
                }

                case IDC_CANCEL:
                    WinSendMsg( hwnd, WM_CLOSE, 0, 0);
                    break;
            }

            return 0;

        case WM_CONTROL:
            if (mp1 == MPFROM2SHORT( IDC_SAVENOW, BN_CLICKED)) {
                USHORT  usChk;

                usChk = (USHORT)(ULONG)WinSendMsg( (HWND)mp2, BM_QUERYCHECK, 0, 0);
                WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVEAFTER), usChk);
                WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVELATER), usChk);
                WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVEDONT), usChk);
            }
            return 0;

        case WM_INITDLG:
            InitSaveDlg( hwnd, mp2);
            return 0;

        case WM_CLOSE: {
            HWND    hOwner = WinQueryWindow( hwnd, QW_OWNER);

            // if mp1 is TRUE, the SaveFiles dialog has been invoked,
            // so don't reenable the main dialog yet
            if (!mp1)
                WinSendMsg( hOwner, CAMMSG_DLGSTATUS, (MP)IDD_SAVE, (MP)FALSE);

            WinDestroyWindow( hwnd);
            return (0);
        }

    } //end switch (msg)

    return (WinDefDlgProc( hwnd, msg, mp1, mp2));
}

/****************************************************************************/

LONG        InitSaveDlg( HWND hwnd, MPARAM mp2)

{
    STATSTUFF *     pss;
    ULONG           cnt;
    ULONG           ulSaveErase;
    HWND            hTemp;
    char            szText[128];
    char            szFmt[128];

    WinSetWindowULong( hwnd, QWL_USER, (ULONG)mp2);

    pss = GetStatStuff();
    LoadDlgStrings( hwnd, nlsSaveDlg);

    hTemp = WinWindowFromID( hwnd, IDC_ERASENOW);
    LoadString( DLG_Erase, szFmt);
    sprintf( szText, szFmt, (int)pss->ulDelCnt, (int)(pss->ulDelSize/1024));
    WinSetWindowText( hTemp, (PCSZ)szText);
    if (pss->ulDelCnt)
        WinSendMsg( hTemp, BM_SETCHECK, (MP)1, 0);
    else
        WinEnableWindow( hTemp, FALSE);

    hTemp = WinWindowFromID( hwnd, IDC_SAVENOW);
    LoadString( DLG_Save, szFmt);
    sprintf( szText, szFmt, (int)pss->ulSavCnt, (int)(pss->ulSavSize/1024));
    WinSetWindowText( hTemp, (PCSZ)szText);

    if (pss->ulSavCnt == 0) {
        WinEnableWindow( hTemp, FALSE);
        WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVEAFTER), FALSE);
        WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVELATER), FALSE);
        WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVEDONT), FALSE);
    }
    else {
        WinSendMsg( hTemp, BM_SETCHECK, (MP)1, 0);

        ulSaveErase = 0;
        cnt = sizeof(ULONG);
        if (!PrfQueryProfileData( HINI_USERPROFILE, CAMINI_APP,
                                  CAMINI_SAVERASE, &ulSaveErase, &cnt) ||
            !cnt || ulSaveErase > 2)
            ulSaveErase = 0;

        if (WinQueryWindowUShort( hwnd, QWS_ID) == IDD_SAVEEXIT) {
            WinEnableWindow( WinWindowFromID( hwnd, IDC_SAVELATER), FALSE);
            if (ulSaveErase == 1)
                ulSaveErase = 0;
        }

        WinSendDlgItemMsg( hwnd, IDC_SAVEOPTS+ulSaveErase, BM_SETCHECK,
                           (MP)1, 0);
    }

    return 0;
}

/****************************************************************************/

BOOL        SaveFiles( HWND hwnd)

{
    ULONG   ulMsg = WinQueryWindowULong( hwnd, QWL_USER);
    ULONG   ulSaveErase = 0;
    HWND    hOwner;
    HWND    hDlg;
    ULONG   flags = 0;

    if (WinSendDlgItemMsg( hwnd, IDC_SAVENOW, BM_QUERYCHECK, 0, 0)) {
        flags |= CAMSAVE_SAVE;

        if (WinSendDlgItemMsg( hwnd, IDC_SAVEAFTER, BM_QUERYCHECK, 0, 0)) {
            flags |= CAMSAVE_SAVEERASE;
            ulSaveErase = 2;
        }
        else
        if (WinSendDlgItemMsg( hwnd, IDC_SAVELATER, BM_QUERYCHECK, 0, 0)) {
            flags |= CAMSAVE_SAVEMARK;
            ulSaveErase = 1;
        }

        PrfWriteProfileData( HINI_USER, CAMINI_APP, CAMINI_SAVERASE,
                             &ulSaveErase, sizeof(ulSaveErase));
    }

    if (WinSendDlgItemMsg( hwnd, IDC_ERASENOW, BM_QUERYCHECK, 0, 0))
        flags |= CAMSAVE_ERASE;

    if (!flags)
        return (FALSE);

    hOwner = WinQueryWindow( hwnd, QW_OWNER);
    hDlg = WinLoadDlg(
                HWND_DESKTOP,                       // parent-window
                hOwner,                             // owner-window
                SaveFilesDlgProc,                   // dialog proc
                NULLHANDLE,                         // EXE module handle
                IDD_SAVEFILE,                       // dialog id
                (PVOID)ulMsg);                      // pointer to create params

    if (hDlg) {
        SetFileSaveSortOrder( hOwner, TRUE);
        WinPostMsg( hDlg, CAMMSG_STARTSAVE, (MP)flags, 0);
    }

    return (hDlg);
}

/****************************************************************************/
/****************************************************************************/

// this is an exercise in non-modality (i.e. no annoying popup windows)

MRESULT _System SaveFilesDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{

    switch (msg)
    {
        case WM_COMMAND: {
            CAMSaveFilePtr  pCSF;

            pCSF = (CAMSaveFilePtr)WinQueryWindowULong( hwnd, QWL_USER);
            if (!pCSF) {
                WinSendMsg( hwnd, WM_CLOSE, 0, 0);
                return 0;
            }

            if (mp2 == (MP)-1) {
                if (SHORT1FROMMP( mp1) == IDC_HALT)
                    pCSF->fHalt = TRUE;
                return 0;
            }

            switch (SHORT1FROMMP( mp1)) {

                case IDC_NO:
                    WinSendMsg( hwnd, WM_CLOSE, 0, 0);
                    break;

                case IDC_NEXT:
                    SaveNextFile( hwnd, pCSF);
                    break;

                case IDC_RETRY:
                    SaveFile( hwnd, pCSF);
                    break;

                case IDC_RETRYDEL:
                    EraseFile( hwnd, pCSF);
                    break;

                case IDC_CREATE:
                    CreateSavePath( hwnd, pCSF);
                    break;

                case IDC_REPLACE:
                    pCSF->flags |= CAMSAVE_REPONE;
                    SaveFile( hwnd, pCSF);
                    break;

                case IDC_STOP:
                    DoneSavingFiles( hwnd, pCSF);
                    break;

                case IDC_HALT:
                    pCSF->fHalt = TRUE;
                    break;
            }
            return (0);
        }

        case CAMMSG_STARTSAVE: {
            CAMSaveFilePtr  pCSF;

            pCSF = (CAMSaveFilePtr)WinQueryWindowULong( hwnd, QWL_USER);
            if (!pCSF) {
                WinSendMsg( hwnd, WM_CLOSE, 0, 0);
                return 0;
            }
            pCSF->flags |= (ULONG)mp1;
            pCSF->pcr = 0;
            SaveNextFile( hwnd, pCSF);
            return 0;
        }

        case WM_CONTROL:
            if (mp1 == MPFROM2SHORT( IDC_REPLACEALL, BN_CLICKED)) {
                CAMSaveFilePtr  pCSF;

                pCSF = (CAMSaveFilePtr)WinQueryWindowULong( hwnd, QWL_USER);
                if (!pCSF) {
                    WinSendMsg( hwnd, WM_CLOSE, 0, 0);
                    return 0;
                }
                if (WinSendMsg( (HWND)mp2, BM_QUERYCHECK, 0, 0))
                    pCSF->flags |= CAMSAVE_REPALL;
                else
                    pCSF->flags &= ~CAMSAVE_REPALL;
            }
            return 0;

        case WM_CLOSE: {
            HWND    hOwner = WinQueryWindow( hwnd, QW_OWNER);

            SetFileSaveSortOrder( hOwner, FALSE);
            // reenable the main dialog
            WinSendMsg( hOwner, CAMMSG_DLGSTATUS, (MP)IDD_SAVE, (MP)FALSE);
            WinDestroyWindow( hwnd);
            return (0);
        }

        case WM_DESTROY: {
            CAMSaveFilePtr  pCSF;
            HWND            hOwner;

            hOwner = WinQueryWindow( hwnd, QW_OWNER);
            pCSF = (CAMSaveFilePtr)WinQueryWindowULong( hwnd, QWL_USER);
            if (pCSF) {
                if (pCSF->ulMsg) {
                    STATSTUFF * pss = GetStatStuff();
                    pss->ulSavCnt = 0;
                    pss->ulSavSize = 0;
                    pss->ulDelCnt = 0;
                    pss->ulDelSize = 0;
                    WinPostMsg( hOwner, LOUSHORT(pCSF->ulMsg),
                                MPFROM2SHORT( HIUSHORT(pCSF->ulMsg), 0), 0);
                }
                free( pCSF);
            }
            break;
        }

        case WM_INITDLG:
            return ((MRESULT)InitSaveFilesDlg( hwnd, mp2));

        case WM_SAVEAPPLICATION:
            WinStoreWindowPos( CAMINI_APP, CAMINI_FILEPOS, hwnd);
            break;

    } //end switch (msg)

    return (WinDefDlgProc( hwnd, msg, mp1, mp2));
}

/****************************************************************************/

LONG        InitSaveFilesDlg( HWND hwnd, MPARAM mp2)

{
    ULONG           ul;
    HWND            hOwner;
    CAMSaveFilePtr  pCSF = 0;

    pCSF = malloc( sizeof(CAMSaveFile));
    if (!pCSF)
        return (-1);

    memset( pCSF, 0, sizeof(CAMSaveFile));
    WinSetWindowULong( hwnd, QWL_USER, (ULONG)pCSF);

    hOwner = WinQueryWindow( hwnd, QW_OWNER);

    pCSF->ulMsg = (ULONG)mp2;
    pCSF->hCnr = WinWindowFromID( hOwner, FID_CLIENT);

    pCSF->hBtn1 = WinWindowFromID( hwnd, IDC_BTN1);
    pCSF->hBtn2 = WinWindowFromID( hwnd, IDC_BTN2);
    pCSF->hBtn3 = WinWindowFromID( hwnd, IDC_BTN3);
    pCSF->hBtn4 = WinWindowFromID( hwnd, IDC_BTN4);
    pCSF->hChk  = WinWindowFromID( hwnd, IDC_REPLACEALL);

    pCSF->hEdit = WinWindowFromID( hwnd, IDC_TEXT2);
    WinSendMsg( pCSF->hEdit, EM_SETTEXTLIMIT, (MP)259, 0);
    if (!WinQueryPresParam( hwnd, PP_BACKGROUNDCOLOR, 0, &ul,
                            sizeof(RGB), &lrgbEditBkgd.rgb, 0))
        lrgbEditBkgd.l = WinQuerySysColor( HWND_DESKTOP,
                                           SYSCLR_DIALOGBACKGROUND, 0);

    if (!fInitSaveFilesMsgs) {
        LoadStaticStrings( nlsSaveFilesMsgs);
        fInitSaveFilesMsgs = TRUE;
    }
    LoadDlgStrings( hwnd, nlsSaveFilesDlg);
    SetSaveControls( hwnd, pCSF, CAMSAVE_RC_INIT);
    WinRestoreWindowPos( CAMINI_APP, CAMINI_FILEPOS, hwnd);

    return (0);
}

/****************************************************************************/

void        SaveNextFile( HWND hwnd, CAMSaveFilePtr pCSF)

{
    ULONG           ulOp;

    ulOp = NextMarkedRec( pCSF->hCnr, &pCSF->pcr, pCSF->flags);

    if (pCSF->pcr) {

        if (pCSF->flags & CAMSAVE_ERRMASK)
            SetSaveControls( hwnd, pCSF, CAMSAVE_RC_OK);

        if (ulOp == CAMSAVE_SAVE)
            SaveFile( hwnd, pCSF);
        else
        if (ulOp == CAMSAVE_ERASE)
            EraseFile( hwnd, pCSF);
    }
    else
        DoneSavingFiles( hwnd, pCSF);

    return;
}

/****************************************************************************/

void        DoneSavingFiles( HWND hwnd, CAMSaveFilePtr pCSF)

{

static BOOL         fRecursion = FALSE;
    HWND            hRect;
    STATSTUFF *     pss;
    char            szText[128];
    char            szFmt[128];

    if (fRecursion)
        printf( "** Recursive entry into DoneSavingFiles **\n");
    else
        fRecursion = TRUE;

    pCSF->flags &= ~CAMSAVE_ERRMASK;
    pCSF->pcr = 0;
    pCSF->fHalt = FALSE;

    pss = GetStatStuff();
    SetSaveControls( hwnd, pCSF, CAMSAVE_RC_DONE);
    hRect = WinWindowFromID( hwnd, IDC_RECT);

    if (pCSF->flags & CAMSAVE_SAVEERASE)
        LoadString( MSG_SaveEraseOK, szFmt);
    else
        LoadString( MSG_SavedOK, szFmt);
    sprintf( szText, szFmt, (int)pCSF->cntSave, (int)(pCSF->cbSave/1024));
    WinSetDlgItemText( hRect, IDC_TEXT1, (PCSZ)szText);

    LoadString( MSG_NotSaved, szFmt);
    sprintf( szText, szFmt, (int)pss->ulSavCnt, (int)(pss->ulSavSize/1024));
    WinSetDlgItemText( hRect, IDC_TEXT2, (PCSZ)szText);

    LoadString( MSG_ErasedOK, szFmt);
    sprintf( szText, szFmt, (int)pCSF->cntErase, (int)(pCSF->cbErase/1024));
    WinSetDlgItemText( hRect, IDC_TEXT3, (PCSZ)szText);

    LoadString( MSG_NotErased, szFmt);
    if (pCSF->flags & CAMSAVE_SAVEMARK)
        sprintf( szText, szFmt, (int)(pss->ulDelCnt - pCSF->cntSave),
                                (int)((pss->ulDelSize - pCSF->cbSave)/1024));
    else
        sprintf( szText, szFmt, (int)pss->ulDelCnt, (int)(pss->ulDelSize/1024));
    WinSetDlgItemText( hRect, IDC_TEXT4, (PCSZ)szText);

    WinShowWindow( hRect, TRUE);
    UpdateStatus( WinQueryWindow( hwnd, QW_OWNER), 0);

    fRecursion = FALSE;

    return;
}

/****************************************************************************/

void        EraseFile( HWND hwnd, CAMSaveFilePtr pCSF)

{
    ULONG           ulErr;
    ULONG           size;
    STATSTUFF *     pss;

do {
    if (pCSF->flags & CAMSAVE_ERRMASK)
        SetSaveControls( hwnd, pCSF, CAMSAVE_RC_OK);

    size = ((CAMRecPtr)pCSF->pcr)->size;
    ulErr = EraseFileInternal( hwnd, pCSF);
    if (ulErr != CAMSAVE_RC_OK)
        break;

    pCSF->cntErase++;
    pCSF->cbErase += size;
    pss = GetStatStuff();
    pss->ulTotCnt--;
    pss->ulTotSize -= size;
    pss->ulDelCnt--;
    pss->ulDelSize -= size;

    WinPostMsg( hwnd, WM_COMMAND, (MP)IDC_NEXT, 0);

} while (0);

    SetSaveControls( hwnd, pCSF, ulErr);

    return;
}

/****************************************************************************/

ULONG       EraseFileInternal( HWND hwnd, CAMSaveFilePtr pCSF)

{
    ULONG           ulErr = CAMSAVE_RC_OK;
    APIRET          rc;
    CAMRecPtr       pcr;
    char            szText[256];

do {
    pcr = pCSF->pcr;

    sprintf( szText, pszMSG_Erasing,
             (pcr->orgname ? pcr->orgname : pcr->title));
    WinSetDlgItemText( hwnd, IDC_TEXT1, szText);
    WinSetDlgItemText( hwnd, IDC_TEXT2, "");

    if (pCSF->fHalt) {
        ulErr = CAMSAVE_RC_CANCELLED;
        break;
    }

    rc = DispatchUsbMsg( CAMMSG_DELETEOBJECT, pcr->hndl, 0);
    if (rc) {
        if (rc == CAMERR_REQCANCELLED)
            ulErr = CAMSAVE_RC_CANCELLED;
        else
            ulErr = CAMSAVE_RC_DELDATA;
        break;
    }

    pCSF->pcr = (CAMRecPtr)WinSendMsg( pCSF->hCnr, CM_QUERYRECORD,
                    (MP)pcr, MPFROM2SHORT( CMA_PREV, CMA_ITEMORDER));

    if (pcr->bmp)
        GpiDeleteBitmap( pcr->bmp);
    if (pcr->orgname)
        free( pcr->orgname);
    WinSendMsg( pCSF->hCnr, CM_REMOVERECORD, (MP)&pcr,
                MPFROM2SHORT( 1, (CMA_FREE | CMA_INVALIDATE)));

} while (0);

    return (ulErr);
}

/****************************************************************************/

void        SaveFile( HWND hwnd, CAMSaveFilePtr pCSF)

{
    HFILE       hFile = 0;
    ULONG       ulErr = CAMSAVE_RC_OK;
    APIRET      rc;
    ULONG       cbPath;
    ULONG       ul;
    char *      ptr;
    STATSTUFF * pss;
    CAMRecPtr   pcr;
    CAMSaveObj  cso;
    FILESTATUS3 fs3;
    EAOP2       eaop2;
    CAMSubjectEA ea;
    char        chSave;
    char        szText[128];

do {
    pcr = pCSF->pcr;

    if (pCSF->flags & CAMSAVE_RETRY) {
        if ((pCSF->flags & CAMSAVE_EDIT) &&
            WinQueryWindowText( pCSF->hEdit, sizeof(pCSF->szPath), pCSF->szPath) == 0)
            cbPath = BuildFilename( pcr, pCSF->szPath);
        else
            cbPath = strlen( pCSF->szPath);

        ul = pCSF->flags & CAMSAVE_REPONE;
        SetSaveControls( hwnd, pCSF, ulErr);
        pCSF->flags |= ul;
    }
    else {
        sprintf( szText, pszMSG_Saving,
                 (pcr->orgname ? pcr->orgname : pcr->title));
        WinSetDlgItemText( hwnd, IDC_TEXT1, szText);
        WinSetWindowUShort( pCSF->hBtn4, QWS_ID, IDC_HALT);

        cbPath = BuildFilename( pcr, pCSF->szPath);
    }

    WinSetWindowText( pCSF->hEdit, pCSF->szPath);

    if (pCSF->fHalt) {
        ulErr = CAMSAVE_RC_CANCELLED;
        break;
    }

    if (cbPath >= CCHMAXPATH) {
        ulErr = CAMSAVE_RC_TOOLONG;
        break;
    }

    rc = DosQueryPathInfo( pCSF->szPath, FIL_STANDARD, &fs3, sizeof(FILESTATUS3));
    if (!rc && !(pCSF->flags & CAMSAVE_REPMASK)) {
        ulErr = CAMSAVE_RC_FILEEXISTS;
        break;
    }

    ptr = strrchr( pCSF->szPath, '\\');
    if (ptr && ptr != pCSF->szPath) {
        if (ptr[-1] == ':')
            ptr++;
        chSave = *ptr;
        *ptr = 0;

        rc = DosQueryPathInfo( pCSF->szPath, FIL_STANDARD, &fs3,
                               sizeof(FILESTATUS3));
        *ptr = chSave;
        if (rc) {
            ulErr = CAMSAVE_RC_BADPATH;
            break;
        }
    }

    BuildSubjectEA( pcr, &ea);
    eaop2.fpGEA2List = 0;
    eaop2.fpFEA2List = (PFEA2LIST)&ea;

    rc = DosOpen( pCSF->szPath, &hFile, &ul, pcr->size, FILE_NORMAL,
                  OPEN_ACTION_CREATE_IF_NEW | ((pCSF->flags & CAMSAVE_REPMASK) ?
                  OPEN_ACTION_REPLACE_IF_EXISTS : OPEN_ACTION_FAIL_IF_EXISTS),
                  (OPEN_FLAGS_FAIL_ON_ERROR | OPEN_FLAGS_SEQUENTIAL |
                  OPEN_SHARE_DENYREADWRITE | OPEN_ACCESS_READWRITE), &eaop2);
    if (rc) {
        ulErr = CAMSAVE_RC_DOSOPEN;
        break;
    }

    cso.hObject = pcr->hndl;
    cso.hFile = hFile;

    rc = DispatchUsbMsg( CAMMSG_SAVEOBJECT, (ULONG)&cso, 0);
    if (rc) {
        if (rc == CAMERR_REQCANCELLED)
            ulErr = CAMSAVE_RC_CANCELLED;
        else
            ulErr = CAMSAVE_RC_GETDATA;
        DosClose( hFile);
        DosDelete( pCSF->szPath);
        break;
    }

    rc = DosQueryFileInfo( hFile, FIL_STANDARD, &fs3, sizeof(FILESTATUS3));
    if (!rc) {
        fs3.fdateCreation.year  = pcr->date.year - 1980;
        fs3.fdateCreation.month = pcr->date.month;
        fs3.fdateCreation.day   = pcr->date.day;

        fs3.ftimeCreation.hours   = pcr->time.hours;
        fs3.ftimeCreation.minutes = pcr->time.minutes;
        fs3.ftimeCreation.twosecs = pcr->time.seconds / 2;

        fs3.fdateLastWrite = fs3.fdateCreation;
        fs3.ftimeLastWrite = fs3.ftimeCreation;

        rc = DosSetFileInfo( hFile, FIL_STANDARD, &fs3, sizeof(FILESTATUS3));
    }

    DosClose( hFile);

    pCSF->cntSave++;
    pCSF->cbSave += pcr->size;

    pss = GetStatStuff();

    if (pCSF->flags & CAMSAVE_SAVEERASE) {
        pss->ulSavCnt--;
        pss->ulSavSize -= pcr->size;
        ulErr = EraseFileInternal( hwnd, pCSF);
        if (ulErr != CAMSAVE_RC_OK)
            break;
    }
    else
    if (pCSF->flags & CAMSAVE_SAVEMARK)
        UpdateMark( pcr, pCSF->hCnr, pss, IDM_SAVEDMARK, FALSE);
    else
        UpdateMark( pcr, pCSF->hCnr, pss, IDM_SAVED, FALSE);

    WinPostMsg( hwnd, WM_COMMAND, (MP)IDC_NEXT, 0);

} while (0);

    SetSaveControls( hwnd, pCSF, ulErr);

    return;
}

/****************************************************************************/

void        SetSaveControls( HWND hwnd, CAMSaveFilePtr pCSF, ULONG ulErr)

{
    BOOL    fEdit = FALSE;
    BOOL    fBtn1 = FALSE;
    BOOL    fChk  = FALSE;
    BOOL    fDel  = FALSE;
    ULONG   ulText3 = 0;
    ULONG   ulText4 = 0;
    char    szText[256];

    pCSF->flags &= ~(CAMSAVE_ERRMASK | CAMSAVE_REPONE);

    if (ulErr == CAMSAVE_RC_INIT) {
        pCSF->flags &= ~CAMSAVE_REPALL;
        WinSetWindowUShort( pCSF->hBtn1, QWS_ID, (USHORT)-1);
        WinSetWindowText( pCSF->hBtn1, "");
        WinSetWindowUShort( pCSF->hBtn2, QWS_ID, IDC_RETRY);
        WinSetWindowText( pCSF->hBtn2, pszBTN_Retry);
        WinSetWindowUShort( pCSF->hBtn3, QWS_ID, IDC_NEXT);
        WinSetWindowText( pCSF->hBtn3, pszBTN_Skip);
        WinSetWindowUShort( pCSF->hBtn4, QWS_ID, IDC_HALT);
        WinSetWindowText( pCSF->hBtn4, pszBTN_Stop);
        ulErr = CAMSAVE_RC_OK;
    }

    if (ulErr == CAMSAVE_RC_OK) {
        WinShowWindow( pCSF->hBtn1, FALSE);
        WinShowWindow( pCSF->hBtn2, FALSE);
        WinShowWindow( pCSF->hBtn3, FALSE);

        WinSendMsg( pCSF->hEdit, EM_SETREADONLY, (MP)TRUE, 0);
        WinSetPresParam( pCSF->hEdit, PP_BACKGROUNDCOLOR,
                         sizeof(RGB), &lrgbEditBkgd.rgb);

        WinSetDlgItemText( hwnd, IDC_TEXT3, "");
        sprintf( szText, pszMSG_SavedErased,
                 (int)pCSF->cntSave, (int)pCSF->cntErase);
        WinSetDlgItemText( hwnd, IDC_TEXT4, szText);
        WinSetWindowUShort( pCSF->hBtn4, QWS_ID, IDC_HALT);
        return;
    }

    if (ulErr == CAMSAVE_RC_DONE) {
        WinShowWindow( pCSF->hBtn1, FALSE);
        WinShowWindow( pCSF->hBtn2, FALSE);
        WinShowWindow( pCSF->hBtn3, FALSE);
        WinShowWindow( pCSF->hChk,  FALSE);
        WinShowWindow( pCSF->hEdit, FALSE);
        WinShowWindow( WinWindowFromID( hwnd, IDC_TEXT1), FALSE);
        WinShowWindow( WinWindowFromID( hwnd, IDC_TEXT3), FALSE);
        WinShowWindow( WinWindowFromID( hwnd, IDC_TEXT4), FALSE);
        WinSetWindowUShort( pCSF->hBtn4, QWS_ID, IDC_NO);
        WinSetWindowText( pCSF->hBtn4, pszBTN_Close);
        WinSendMsg( pCSF->hBtn4, BM_SETDEFAULT, (MP)TRUE, 0);
        return;
    }

    switch (ulErr) {

        case CAMSAVE_RC_TOOLONG: {
            ulText3 = MSG_TooLong1;
            ulText4 = MSG_TooLong2;
            fEdit = TRUE;
            break;
        }

        case CAMSAVE_RC_FILEEXISTS: {
            ulText3 = MSG_FileExists1;
            ulText4 = MSG_FileExists2;
            WinSetWindowUShort( pCSF->hBtn1, QWS_ID, IDC_REPLACE);
            WinSetWindowText( pCSF->hBtn1, pszBTN_Replace);
            fEdit = TRUE;
            fBtn1 = TRUE;
            fChk  = TRUE;
            break;
        }

        case CAMSAVE_RC_BADPATH: {
            ulText3 = MSG_PathNotFound1;
            ulText4 = MSG_PathNotFound2;
            WinSetWindowUShort( pCSF->hBtn1, QWS_ID, IDC_CREATE);
            WinSetWindowText( pCSF->hBtn1, pszBTN_Create);
            fEdit = TRUE;
            fBtn1 = TRUE;
            break;
        }

        case CAMSAVE_RC_BADCREATE: {
            ulText3 = MSG_CouldNotCreate1;
            ulText4 = MSG_CouldNotCreate2;
            WinSetWindowUShort( pCSF->hBtn1, QWS_ID, IDC_CREATE);
            WinSetWindowText( pCSF->hBtn1, pszBTN_Create);
            fEdit = TRUE;
            fBtn1 = TRUE;
            break;
        }

        case CAMSAVE_RC_GETDATA: {
            ulText3 = MSG_GetData;
            break;
        }

        case CAMSAVE_RC_CANCELLED: {
            pCSF->fHalt = FALSE;
            ulText3 = MSG_OpCancelled;
            break;
        }

        case CAMSAVE_RC_DOSOPEN: {
            ulText3 = MSG_OpenFile;
            fEdit = TRUE;
            break;
        }

        case CAMSAVE_RC_DOSWRITE: {
            ulText3 = MSG_WriteFile;
            fEdit = TRUE;
            break;
        }

        case CAMSAVE_RC_DELDATA: {
            ulText3 = MSG_ErasePhoto;
            fDel = TRUE;
            break;
        }

        default:
            return;
    }

    if (ulText3)
        LoadString( ulText3, szText);
    else
        szText[0] = 0;
    WinSetDlgItemText( hwnd, IDC_TEXT3, szText);

    if (ulText4)
        LoadString( ulText4, szText);
    else
        szText[0] = 0;
    WinSetDlgItemText( hwnd, IDC_TEXT4, szText);

    WinShowWindow( pCSF->hBtn1, fBtn1);
    WinShowWindow( pCSF->hBtn2, TRUE);
    WinShowWindow( pCSF->hBtn3, TRUE);
    WinShowWindow( pCSF->hChk,  fChk);

    WinSetWindowUShort( pCSF->hBtn2, QWS_ID,
                        (fDel ? IDC_RETRYDEL : IDC_RETRY));
    WinSetWindowUShort( pCSF->hBtn4, QWS_ID, IDC_STOP);

    WinSendMsg( pCSF->hEdit, EM_SETREADONLY, (MP)(fEdit ? FALSE : TRUE), 0);
    WinSetPresParam( pCSF->hEdit, PP_BACKGROUNDCOLOR,
                     sizeof(RGB), (fEdit ? &lrgbWhite.rgb : &lrgbEditBkgd.rgb));

    pCSF->flags |= CAMSAVE_RETRY | (fEdit ? CAMSAVE_EDIT : 0);

    return;
}

/****************************************************************************/

ULONG       BuildFilename( CAMRecPtr pcr, char * pszName)
{

    BOOL        fNbr;
    CAMGroupPtr pcg;
    char *  	pOut;
    char *      pTmpl;
    char *      pTmplName;
    char *      pTmplExt;
    char *      pFile;
    char *      pFileExt;
    char *      pEnd;
    char *      ptr;

do {
    *pszName = 0;
    pOut = pszName;

    pcg = (CAMGroupPtr)pcr->grpptr;

    // if the title has been changed, use it as a template;
    // otherwise, use the group's filename template;
    // in either case, pFile points at the original filename
    if (pcr->orgname) {
        pTmpl = pcr->title;
        pFile = pcr->orgname;
    }
    else {
        pTmpl = pcg->file;
        pFile = pcr->title;
    }

    // see if the template includes a path
    // (this should only be true if the title was changed)
    pTmplName = strrchr( pTmpl, '\\');
    if (!pTmplName)
        pTmplName = strchr( pTmpl, ':');

    // if there's a path, use it, else use the group's path if present
    if (pTmplName) {
        memcpy( pOut, pTmpl, (pTmplName - pTmpl + 1));
        pOut += (pTmplName - pTmpl + 1);
    }
    else
    if (*pcg->path) {
        strcpy( pOut, pcg->path);
        pOut = strchr( pOut, 0);
        if (pOut[-1] != '\\')
            *pOut++ = '\\';
    }

    // point at the beginning of the name part of the template
    if (pTmplName)
        pTmplName++;
    else
        pTmplName = pTmpl;

    // a quick out for simple cases
    if (pTmplName[0] == 0 ||
        (pTmplName[0] == '*' && pTmplName[1] == 0) ||
        !strcmp( pTmplName, "*.*")) {
        strcpy( pOut, pFile);
        pOut = strchr( pOut, 0);
        break;
    }

    // locate the template's extension
    pTmplExt = strrchr( pTmplName, '.');
    if (pTmplExt)
        pEnd = pTmplExt;
    else
        pEnd = strchr( pTmplName, 0);

    // locate the file's extension
    pFileExt = strrchr( pFile, '.');

    // transfer the name portion of the template
    for (ptr = pTmplName, fNbr = FALSE; ptr < pEnd; ptr++) {

        // insert the original filename
        if (*ptr == '*') {
            if (pFileExt) {
                memcpy( pOut, pFile, pFileExt - pFile);
                pOut += pFileExt - pFile;
            }
            else {
                strcpy( pOut, pFile);
                pOut = strchr( pOut, 0);
            }
        }
        else
        // insert a sequence nbr & signal it needs to be incremented later
        if (*ptr == '#') {
            fNbr = TRUE;
            pOut += sprintf( pOut, "%0*d", (int)pcg->cntDigits, (int)pcg->seqNbr);
        }
        else
        // just copy the character
            *pOut++ = *ptr;
    }
    *pOut = 0;

    // increment the sequence nbr after the loop
    // to keep it from changing within a filename
    if (fNbr)
        pcg->seqNbr++;

    // if the template doesn't have an extension,
    // use the original file's if present, then exit
    if (!pTmplExt) {
        if (pFileExt) {
            strcpy( pOut, pFileExt);
            pOut = strchr( pOut, 0);
        }
        break;
    }

    // if there's a file extension, point it at the char after the dot
    if (pFileExt)
        pFileExt++;

    // transfer the template's extension
    for (ptr = pTmplExt; *ptr; ptr++) {
        // insert the original file's extension if there is one
        if (*ptr == '*') {
            if (pFileExt) {
                strcpy( pOut, pFileExt);
                pOut = strchr( pOut, 0);
            }
        }
        else
        // just copy the character
            *pOut++ = *ptr;
    }
    *pOut = 0;

} while (0);

    // remove CR & LF
    ptr = pszName;
    while ((ptr = strpbrk( ptr, "\r\n")) != 0)
        strcpy( ptr, ptr + strspn( ptr, "\r\n"));

    return (strlen( pszName));
}

/****************************************************************************/

// prepares the file's original name to be written as its
// .SUBJECT extended attribute

void        BuildSubjectEA( CAMRecPtr pcr, CAMSubjectEA * pEA)

{
    ULONG       cbName;
    char *      pszName;

    pszName = (char*)(pcr->orgname ? pcr->orgname : pcr->title);
    cbName = strlen( pszName);

    pEA->hdr.ulcbList = sizeof(pEA->hdr) + sizeof(pEA->info) + cbName;
    pEA->hdr.uloNextEntryOffset = 0;
    pEA->hdr.bfEA = 0;
    pEA->hdr.bcbName = sizeof(".SUBJECT") - 1;
    pEA->hdr.uscbValue = sizeof(pEA->info) + cbName;
    strcpy(pEA->hdr.chszName, ".SUBJECT");
    pEA->info.usEAType = EAT_ASCII;
    pEA->info.usDataLth = cbName;
    strcpy( pEA->data, pszName);

    return;
}

/****************************************************************************/

void        CreateSavePath( HWND hwnd, CAMSaveFilePtr pCSF)

{
    APIRET      rc;
    char *      ptr;
    char        szPath[CCHMAXPATH];
    char        szFile[CCHMAXPATH];

do {

    if (WinQueryWindowText( pCSF->hEdit, sizeof(szPath), szPath) == 0) {
        WinSetWindowText( pCSF->hEdit, pCSF->szPath);
        SetSaveControls( hwnd, pCSF, CAMSAVE_RC_BADCREATE);
        break;
    }

    ptr = strrchr( szPath, '\\');
    if (ptr) {
        strcpy( szFile, ptr+1);
        if (ptr == szPath || ptr[-1] == ':')
            ptr++;
        *ptr = 0;
    }
    else {
        strcpy( szFile, szPath);
        strcpy( szPath, ".");
    }

    rc = CreatePath( szPath);
    ptr = strchr( szPath, 0);
    if (ptr[-1] != '\\')
        *ptr++ = '\\';
    strcpy( ptr, szFile);
    WinSetWindowText( pCSF->hEdit, szPath);

    if (rc) {
        SetSaveControls( hwnd, pCSF, CAMSAVE_RC_BADCREATE);
        break;
    }

    SaveFile( hwnd, pCSF);

} while (0);

    return;

}

/****************************************************************************/

ULONG       NextMarkedRec( HWND hCnr, CAMRecPtr * ppcrIn, ULONG flags)

{
    ULONG       ulRtn = 0;
    ULONG       status;
    CAMRecPtr   pcrOut = *ppcrIn;

    pcrOut = (CAMRecPtr)WinSendMsg(
                 hCnr, CM_QUERYRECORD, (MP)pcrOut,
                 MPFROM2SHORT( (pcrOut ? CMA_NEXT : CMA_FIRST), CMA_ITEMORDER));

    while (pcrOut) {
        if (pcrOut == (CAMRecPtr)-1) {
            pcrOut = 0;
            break;
        }

        status = pcrOut->status;

        if ((flags & CAMSAVE_SAVE) &&
            (status & (STAT_SAVE | STAT_DONE)) == STAT_SAVE) {
            ulRtn = CAMSAVE_SAVE;
            break;
        }

        if ((flags & CAMSAVE_ERASE) && (status & STAT_DELETE)) {
            ulRtn = CAMSAVE_ERASE;
            break;
        }

        pcrOut = (CAMRecPtr)WinSendMsg( hCnr, CM_QUERYRECORD,
                    (MP)pcrOut,  MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER));
    }

    *ppcrIn = pcrOut;

    return (ulRtn);
}

/****************************************************************************/
/****************************************************************************/

ULONG       GetData( HWND hwnd, BOOL fSync)

{
    ULONG           rc = 0;
    HWND            hCnr;
    HWND            hWait;
    STATSTUFF *     pss;
    CAMObjectHandlePtr  pHandles = 0;
    CAMRecPtr       pcr;
    CNRINFO         cnri;
    ULONG           ulErr = STS_Copyright;

do
{
    // show the "please wait" dialog
    hWait = OpenWaitOrExitDlg( hwnd, IDD_WAIT);

    // init the counters
    pss = GetStatStuff();
    pss->ulTotCnt = 0;
    pss->ulTotSize = 0;
    pss->ulSavCnt = 0;
    pss->ulSavSize = 0;
    pss->ulDelCnt = 0;
    pss->ulDelSize = 0;

    if (!fInitColMsgs) {
        LoadStaticStrings( nlsColMsgs);
        fInitColMsgs = TRUE;
    }

    hCnr = WinWindowFromID( hwnd, FID_CLIENT);
    pcr = 0;
    NextRec( hCnr, &pcr);

    // if there aren't any records, use GetList instead of SyncList;
    // otherwise, empty the container if GetList was requested
    if (!pcr)
        fSync = FALSE;
    else
        if (!fSync) {
            while (pcr) {
                if (pcr->bmp)
                    GpiDeleteBitmap( pcr->bmp);
                if (pcr->orgname) 
                    free( pcr->orgname);

                NextRec( hCnr, &pcr);
            }

            if (WinSendMsg( hCnr, CM_REMOVERECORD, 0,
                            MPFROM2SHORT( 0, CMA_FREE | CMA_INVALIDATE)) != 0)
                printf( "GetData - CM_REMOVERECORD failed\n");
        }

    // set the sort function to be used when inserting records
    cnri.cb = sizeof(CNRINFO);
    cnri.pSortRecord = QuerySortFunctionPtr();
    WinSendMsg( hCnr, CM_SETCNRINFO, (MP)&cnri, (MP)CMA_PSORTRECORD);

    if (CAM_DEMO_FLAG) {
        InitColumnWidths( hCnr);
        break;
    }

    // get a list of handles
    rc = DispatchUsbMsg( CAMMSG_GETOBJECTHANDLES, 0, (void**)&pHandles);
    if (rc) {
        printf( "DispatchUsbMsg #1 - rc= %x\n", (int)rc);
        DosSleep( 500);
        rc = DispatchUsbMsg( CAMMSG_GETOBJECTHANDLES, 0, (void**)&pHandles);
        if (rc) {
            printf( "DispatchUsbMsg #2 - rc= %x\n", (int)rc);
            ulErr = STS_GetList;
            break;
        }
    }

    // Dave's fix
    if (CAM_DAVE_FLAG)
        AdjustList( pHandles);

    if (fSync) {
        rc = SyncList( hCnr, pHandles);
        if (rc) {
            ulErr = STS_RefreshList;
            break;
        }
    }
    else {
        rc = GetList( hCnr, pHandles);
        if (rc) {
            if (rc == 4)
                ulErr = STS_NoFiles;
            else
                ulErr = STS_ListingFiles;
            break;
        }
        // if we're in details view, this set the column widths & splitbar
        InitColumnWidths( hCnr);
    }

    WinPostMsg( hwnd, CAMMSG_FETCHTHUMBS, 0, 0);

} while (0);

    if (pHandles)
        free( pHandles);

    // get rid of the wait dialog, then update the status bar
    CloseWaitOrExitDlg( hWait);
    UpdateStatus( hwnd, ulErr);

    return (rc);
}

/****************************************************************************/

// The list of file handles generated by Dave M's camera had invalid entries
// at the beginning, followed by valid ones.  The invalid ones could be
// identified because their numbers were out of sequence.  This works around
// the issue by starting at the end of the list and ensuring that each
// preceeding entry has a lower handle number.  When it finds an out-of-
// sequence entry, it assumes that everything from that point back to the
// beginning is invalid and adjusts the list pointer & count accordingly.

void        AdjustList( CAMObjectHandlePtr pHandles)

{
    ULONG               ctr;
    ULONG               last;
    PULONG              pHndl;

    ctr = pHandles->cntHandles;
    pHndl = (ULONG*)pHandles->Handles;
    last = ~0;

    while (ctr) {
        ctr--;
        if (pHndl[ctr] >= last) {
            ctr++;
            break;
        }
        last = pHndl[ctr];
    }

    if (ctr) {
        printf( "AdjustList - ignoring first %d of %d files\n",
                (int)ctr, (int)pHandles->cntHandles);
        pHandles->Handles = (uint32_t*)&pHndl[ctr];
        pHandles->cntHandles -= ctr;
    }

    return;
}

/****************************************************************************/

ULONG       GetList( HWND hCnr, CAMObjectHandlePtr pHandles)

{
    ULONG               rc = 0;
    ULONG               ctr;
    ULONG               cnt;
    ULONG               cntItems = 0;
    ULONG               cntFldrs = 0;
    PULONG              pHndl;
    CAMObjectInfoPtr    pObjInfo = 0;
    CAMRecPtr           pcr;
    CAMRecPtr           pcrSav;
    CAMRecPtr           pcrFirst = 0;
    RECORDINSERT        ri;

do
{
    cnt = pHandles->cntHandles;
    if (!cnt)
        ERRNBR( 6)

    pHndl = (ULONG*)pHandles->Handles;

    // allocate a record for each item
    pcrFirst = (CAMRecPtr)WinSendMsg( hCnr, CM_ALLOCRECORD,
                                      (MP)CAMREC_EXTRABYTES, (MP)cnt);
    if (pcrFirst == 0)
        ERRNBR( 1)

    // for each handle, get its info then format a record with its data
    for (ctr=0, pcr=pcrFirst, pcrSav=0; pcr && ctr < cnt; ctr++, pHndl++) {

        if (DispatchUsbMsg( CAMMSG_GETOBJECTINFO, *pHndl, (void**)&pObjInfo))
            ERRNBR( 2)

        // skip folders
        if (pObjInfo->ObjectFormat == PTP_OFC_Association) {
            free( pObjInfo);
            cntFldrs++;
            pObjInfo = 0;
            continue;
        }

        cntItems++;
        SetData( pcr, *pHndl, pObjInfo);
        free( pObjInfo);
        pObjInfo = 0;

        pcrSav = pcr;
        pcr = (CAMRecPtr)pcr->mrc.preccNextRecord;
    }

    // having nothing to insert may be an error
    if (cntItems == 0) {
        if (cntFldrs == 0)
            rc = 3;
        else
            rc = 4;
    }

    // if there weren't any errors, insert the records
    if (!rc) {
        // break the linked list at pcrSav which was the last record used
        pcrSav->mrc.preccNextRecord = 0;

        // init the recordinsert struct
        ri.cb = sizeof(RECORDINSERT);
        ri.pRecordOrder = (PRECORDCORE)CMA_FIRST;
        ri.pRecordParent = 0;
        ri.fInvalidateRecord = TRUE;
        ri.zOrder = CMA_TOP;
        ri.cRecordsInsert = cntItems;

        // insert all the records at once;  if this fails, relink the list
        if (!WinSendMsg( hCnr, CM_INSERTRECORD, (MP)pcrFirst, (MP)&ri)) {
            pcrSav->mrc.preccNextRecord = (CAM_RECTYPE*)pcr;
            rc = 5;
        }
    }

    // if there were any errors, set up to delete every record
    if (rc)
        pcr = pcrFirst;

    // free any records that weren't used;  normally,
    // these reflect folder entries that weren't listed
    while (pcr) {
        pcrSav = (CAMRecPtr)pcr->mrc.preccNextRecord;
        if (!WinSendMsg( hCnr, CM_FREERECORD, (MP)&pcr, (MP)1))
            printf( "GetList:  failed to free a container record\n");
        pcr = pcrSav;
    }

} while (0);

    free( pObjInfo);

    if (rc) {
        if (rc == 2)
            printf( "GetList:  GetObjectInfo - handle= %d\n", (int)*pHndl);
        else
            printf( "GetList:  rc= %d\n", (int)rc);
    }

    return (rc);
}

/****************************************************************************/

ULONG       SyncList( HWND hCnr, CAMObjectHandlePtr pHandles)

{
    ULONG               cntFirst = 0;

    ULONG               rc = 0;
    ULONG               ctr;
    ULONG               cnt;
    PULONG              pHndl;
    STATSTUFF *         pss;
    CAMObjectInfoPtr    pObjInfo = 0;
    CAMRecPtr           pcr;
    CAMRecPtr           pcrNext;
    CAMRecPtr           pcrFirst;
    RECORDINSERT        ri;

do
{
    cnt = pHandles->cntHandles;
    pcrFirst = 0;
    NextRec( hCnr, &pcrFirst);

    // for each handle, find a record that matches & flag it;
    // if a match can't be found, insert a new record
    for (ctr=0, pHndl=(ULONG*)pHandles->Handles; ctr < cnt; ctr++, pHndl++) {

        if (pObjInfo) {
            free( pObjInfo);
            pObjInfo = 0;
        }

        if (DispatchUsbMsg( CAMMSG_GETOBJECTINFO, *pHndl, (void**)&pObjInfo)) {
            printf( "Could not get object info for handle %d\n", (int)*pHndl);
            rc = 1;
            break;
        }

        // skip folders
        if (pObjInfo->ObjectFormat == PTP_OFC_Association)
            continue;

        // scan the records for a match;  if found, flag it and
        // update its handle (other details should not have changed)
        pcr = pcrFirst;
        while (pcr) {
            if (!(pcr->status & STAT_MATCHED) &&
                !strcmp( pObjInfo->Filename, (pcr->orgname ?
                                              pcr->orgname : pcr->title))) {
                pcr->status |= STAT_MATCHED;
                pcr->hndl = *pHndl;
                if (pcr == pcrFirst) {
                    cntFirst++;
                    NextRec( hCnr, &pcrFirst);
                }
                break;
            }

            NextRec( hCnr, &pcr);
        }
        if (pcr)
            continue;

        // if not found, insert a new record
        pcr = (CAMRecPtr)WinSendMsg( hCnr, CM_ALLOCRECORD,
                                     (MP)CAMREC_EXTRABYTES, (MP)1);
        if (!pcr)
            ERRNBR( 2)

        SetData( pcr, *pHndl, pObjInfo);
        pcr->status |= STAT_MATCHED;

        ri.cb = sizeof(RECORDINSERT);
        ri.pRecordOrder = (PRECORDCORE)CMA_END;
        ri.pRecordParent = 0;
        ri.fInvalidateRecord = TRUE;
        ri.zOrder = CMA_TOP;
        ri.cRecordsInsert = 1;
        if (!WinSendMsg( hCnr, CM_INSERTRECORD, (MP)pcr, (MP)&ri)) {
            WinSendMsg( hCnr, CM_FREERECORD, (MP)&pcr, (MP)1);
            ERRNBR( 3)
        }

    } // end for()

    if (rc)
        break;

    // delete records that weren't matched & recalc totals
    pss = GetStatStuff();
    pcr = 0;
    NextRec( hCnr, &pcr);

    while (pcr) {
        pcrNext = pcr;
        NextRec( hCnr, &pcrNext);

        if (pcr->status & STAT_MATCHED) {
            pcr->status &= ~STAT_MATCHED;
            pss->ulTotCnt++;
            pss->ulTotSize += pcr->size;
            if (STAT_NOTSAV(pcr->status)) {
                pss->ulSavCnt++;
                pss->ulSavSize += pcr->size;
            }
            else
            if (STAT_ISDEL(pcr->status)) {
                pss->ulDelCnt++;
                pss->ulDelSize += pcr->size;
            }
        }
        else {
            if (pcr->bmp)
                GpiDeleteBitmap( pcr->bmp);
            if (pcr->orgname)
                free( pcr->orgname);
            WinSendMsg( hCnr, CM_REMOVERECORD, (MP)&pcr, MPFROM2SHORT( 1, CMA_FREE));
        }

        pcr = pcrNext;
    }

} while (0);

    free( pObjInfo);

    if (rc)
        printf( "SyncList:  rc= %d\n", (int)rc);

    return (rc);
}

/****************************************************************************/

void        SetData( CAMRecPtr pcr, ULONG ulHndl, CAMObjectInfoPtr pObjInfo)

{
    STATSTUFF *     pss;
    struct tm *     ptm;
    char            szText[32];

    pss = GetStatStuff();
    pss->ulTotCnt++;
    pss->ulTotSize += pObjInfo->ObjectCompressedSize;

    // mark & status
    pcr->mark = pss->hBlkIco;
    pcr->status = 0;

    // number & handle
    pcr->nbr = pss->ulTotCnt;
    pcr->hndl = ulHndl;

    // title
    pcr->title = pcr->sztitle;
    strcpy( pcr->sztitle, pObjInfo->Filename);
    pcr->orgname = 0;

    // group
    pcr->grpptr = GetDefaultGroup( &pcr->group);

    // image format
    pcr->fmt = GetFormatName( pObjInfo->ObjectFormat);
    pcr->fmtnbr = pObjInfo->ObjectFormat;

    // image XY
    pcr->picx = pObjInfo->ImagePixWidth;
    pcr->picy = pObjInfo->ImagePixHeight;
    if (sprintf( szText, "%dx%d",
        pObjInfo->ImagePixWidth, pObjInfo->ImagePixHeight) > 11)
        strcpy( szText, pszCOL_TooBig);
    pcr->picxy = pcr->szpicxy;
    strcpy( pcr->szpicxy, szText);

    // thumb format
    pcr->tnfmt = GetFormatName( pObjInfo->ThumbFormat);
    pcr->tnfmtnbr = pObjInfo->ThumbFormat;

    // thumb XY
    pcr->tnx = pObjInfo->ThumbPixWidth;
    pcr->tny = pObjInfo->ThumbPixHeight;
    if (sprintf( szText, "%dx%d",
        pObjInfo->ThumbPixWidth, pObjInfo->ThumbPixHeight) > 11)
        strcpy( szText, pszCOL_TooBig);
    pcr->tnxy = pcr->sztnxy;
    strcpy( pcr->sztnxy, szText);

    // thumb size
    pcr->tnsize = pObjInfo->ThumbCompressedSize;

    // image size
    pcr->size = pObjInfo->ObjectCompressedSize;

    ptm = localtime( &pObjInfo->CaptureDate);

    // day
    pcr->day = aszDay[ptm->tm_wday];

    // date
    pcr->date.day   = (UCHAR)ptm->tm_mday;
    pcr->date.month = (UCHAR)ptm->tm_mon+1;
    pcr->date.year  = (USHORT)ptm->tm_year+1900;

    // time
    pcr->time.hours   = ptm->tm_hour;
    pcr->time.minutes = ptm->tm_min;
    pcr->time.seconds = ptm->tm_sec;

    // icon view info - the pszIcon field of the (MINI)RECORDCORE
    // has to have something or the container won't allocate space
    // for a title when in icon view;  the icon or bitmap field can
    // be empty because the container is ownerdraw in icon view
    pcr->mrc.pszIcon = pcr->sztitle;
    pcr->bmp = 0;

    return;
}

/****************************************************************************/

void        NextRec( HWND hCnr, CAMRecPtr * ppcrIn)

{
    *ppcrIn = (CAMRecPtr)WinSendMsg( hCnr, CM_QUERYRECORD, (MP)*ppcrIn,
               MPFROM2SHORT( (*ppcrIn ? CMA_NEXT : CMA_FIRST), CMA_ITEMORDER));

    if (*ppcrIn == (CAMRecPtr)-1)
        *ppcrIn = 0;

    return;
}

/****************************************************************************/

