/****************************************************************************/
// cammain.c
//
//  - main(), the main window's dialog proc, & top-level init stuff
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#include "camapp.h"

/****************************************************************************/
/*
void            UpdateStatus( HWND hwnd, ULONG ulMsg);
STATSTUFF *     GetStatStuff( void);
ULONG           ShowInitErrorDlg( ULONG ulErr, ULONG ulDlgID);
HWND            OpenWaitOrExitDlg( HWND hOwner, ULONG ulDlgID);
void            CloseWaitOrExitDlg( HWND hDlg);
*/

void            Morph( void);
void            CheckArgs( int argc, char * argv[]);
MRESULT _System MainWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
void            InitDlg( HWND hwnd);
MRESULT         FormatFrame( HWND hwnd, MPARAM mp1, MPARAM mp2);
BOOL            WM_CHAR_Handler( HWND hwnd, USHORT usVK);
void            InitStatStuff( HWND hwnd);
void            DialogStatus( HWND hwnd, ULONG ulID, BOOL fOpen);

MRESULT _System InitErrorDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT _System WaitOrExitDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);

/****************************************************************************/

char    szCopyright[] = "Cameraderie - (C)Copyright 2006  R.L.Walsh";

// declared EXTERN in camapp.h - contains commandline flags
ULONG   ulGlobalFlags = 0;

// static variables
static BOOL         fRestoreDefaults = FALSE;
static BOOL         fInSave = FALSE;
static HPOINTER     hAppIco = 0;
static STATSTUFF    ss;

static CAMNLS   nlsNoCamDlg[] = {
    { IDC_TEXT1,        DLG_NoCameras },
    { IDC_TEXT2,        DLG_PlugIn },
    { IDC_TEXT4,    	DLG_ToEndNow },
    { IDC_RETRY,    	BTN_Retry },
    { IDC_EXIT,     	BTN_Exit },
    { 0,                0 }};

static CAMNLS   nlsErrorDlg[] = {
    { IDC_TEXT4,    	DLG_ToEnd },
    { IDC_EXIT,     	BTN_Exit },
    { 0,                0 }};

static CAMNLS   nlsLoadingDlg[] = {
    { IDC_TEXT1,        DLG_Loading },
    { IDC_TEXT4,    	DLG_ToEndNow },
    { IDC_EXIT,     	BTN_Exit },
    { 0,                0 }};

static CAMNLS   nlsDisconnectDlg[] = {
    { IDC_DISCON1,      DLG_Disconnected1 },
    { IDC_DISCON2,      DLG_Disconnected2 },
    { IDC_DISCON3,      DLG_Disconnected3 },
    { IDC_DISCON4,      DLG_WindowWillClose1 },
    { IDC_DISCON5,      DLG_WindowWillClose2 },
    { IDC_DISCON6,      DLG_ToEndNow },
    { IDC_EXIT,         BTN_Exit },
    { 0,                0 }};

/****************************************************************************/

int         main( int argc, char * argv[])

{
    HAB     hab = 0;
    HMQ     hmq = 0;
    HWND    hwnd = 0;
    int     nRtn = 8;
    QMSG    qmsg;

do
{
    Morph();
    CheckArgs( argc, argv);

    hab = WinInitialize( 0);
    if (!hab)
        break;

    hmq = WinCreateMsgQueue( hab, 0);
    if (!hmq)
        break;

    LoadNLS();

    if (!CheckVersion()) {
        nRtn = 0;
        break;
    }

    if (!CAM_DEMO_FLAG)
        if (SetupCamera( hmq)) {
            nRtn = 0;
            break;
        }

    hwnd = WinLoadDlg(
                HWND_DESKTOP,               //  parent-window
                NULLHANDLE,                 //  owner-window
                MainWndProc,                //  dialog proc
                NULLHANDLE,                 //  EXE module handle
                IDD_MAIN,                   //  dialog id
                NULL);                      //  pointer to create params

    if (!hwnd)
        break;

    while (WinGetMsg( hab, &qmsg, NULLHANDLE, 0, 0))
        WinDispatchMsg( hab, &qmsg);

    nRtn = 0;

} while (0);

    if (nRtn)
        DosBeep( 440, 150);

    if (hwnd)
        WinDestroyWindow( hwnd);

    ShutDownCamera();

    if (hmq)
        WinDestroyMsgQueue( hmq);
    if (hab)
        WinTerminate( hab);

    return (nRtn);
}

/****************************************************************************/

// enable this program to use PM functions
// even if it was linked or started as a VIO app

void        Morph( void)

{
    PPIB    ppib;
    PTIB    ptib;

    DosGetInfoBlocks( &ptib, &ppib);
    ppib->pib_ultype = 3;
    return;
}

/****************************************************************************/

// currently, this only looks for "demo", "force", & "dave";
// args can be preceeded by '/', '-', '--', or nothing

void        CheckArgs( int argc, char * argv[])

{
    int     ctr;
    char *  ptr;

    ulGlobalFlags = 0;

    if (argc < 2)
        return;

    for (ctr = 1; ctr < argc; ctr++) {
        ptr = argv[ctr];
        if (*ptr == '/')
            ptr++;
        else
        if (*ptr == '-') {
            ptr++;
            if (*ptr == '-')
                ptr++;
        }

        // if we're in demo mode, no other args are meaningful
        if (!stricmp( ptr, "demo")) {
            ulGlobalFlags |= CAMFLAG_DEMO;
            printf( "Cameraderie is in Demo mode\n");
            break;
        }

        // force the camera discovery routine to query devices
        // that don't identify themselves as PTP-class devices
        if (!stricmp( ptr, "force"))
            ulGlobalFlags |= CAMFLAG_FORCE;
        else
        // ignore handles at the beginning of handle list
        // that are out of sequence
        if (!stricmp( ptr, "dave"))
            ulGlobalFlags |= CAMFLAG_DAVE;
    }

    return;
}

/****************************************************************************/
/****************************************************************************/

// just your basic dialog procedure

MRESULT _System MainWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{
    switch (msg)
    {

        case WM_COMMAND:
            Command( hwnd, SHORT1FROMMP( mp1));
            return (0);

        case WM_CONTROL:
            switch ((ULONG)mp1) {
                case MAKEULONG( FID_CLIENT, CN_CONTEXTMENU):
                    PopupMenu( hwnd, (CAMRecPtr)mp2);
                    break;

                case MAKEULONG( FID_CLIENT, CN_ENTER):
                    if (((PNOTIFYRECORDENTER)mp2)->pRecord)
                        Command( hwnd, IDM_WRITE);
                    break;

                case MAKEULONG( FID_CLIENT, CN_EMPHASIS): {
                    PNOTIFYRECORDEMPHASIS pnre = (PNOTIFYRECORDEMPHASIS)mp2;

                    if (!fInSave &&
                        (pnre->fEmphasisMask & CRA_CURSORED) &&
                        (pnre->pRecord->flRecordAttr & CRA_CURSORED))
                        WinPostMsg( hwnd, CAMMSG_SHOWTHUMB,
                                    (MP)pnre->pRecord, 0);
                    break;
                }

                case MAKEULONG( FID_CLIENT, CN_BEGINEDIT): {
                    PCNREDITDATA    pCE = (PCNREDITDATA)mp2;
                    CAMRecPtr       pcr = (CAMRecPtr)pCE->pRecord;

                    if (pcr && pcr->orgname == 0)
                        pcr->orgname = strdup( pcr->title);
                    break;
                }

                case MAKEULONG( FID_CLIENT, CN_REALLOCPSZ): {
                    PCNREDITDATA    pCE = (PCNREDITDATA)mp2;

                    if (pCE->pRecord)
                        if (pCE->cbText > 1 && pCE->cbText <= 260)
                            return (MRESULT)TRUE;
                    break;
                }

                case MAKEULONG( FID_CLIENT, CN_ENDEDIT): {
                    PCNREDITDATA    pCE = (PCNREDITDATA)mp2;
                    CAMRecPtr       pcr = (CAMRecPtr)pCE->pRecord;

                    if (!pcr)
                        break;

                    if (pcr->orgname &&
                        !strcmp( pcr->orgname, pcr->title)) {
                        free( pcr->orgname);
                        pcr->orgname = 0;
                    }
                    else
                        if (IsMultipleSel( hwnd, pcr))
                            SetMultTitles( hwnd, pcr);

                    pcr = (CAMRecPtr)WinSendMsg( pCE->hwndCnr,
                                                 CM_QUERYRECORDEMPHASIS,
                                                 (MP)CMA_FIRST, (MP)CRA_CURSORED);
                    if (pcr)
                        SetThumbWndTitle( hwnd, pcr->title);

                    break;
                }

            }
            return (0);

        case WM_DRAWITEM:
            return ((MRESULT)CnrPaintRecord( (POWNERITEM)mp2));

        case CAMMSG_SHOWTHUMB:
            SetThumbWnd( hwnd, (CAMRecPtr)mp1);
            return (0);

        case CAMMSG_FETCHTHUMBS:
            FetchAllThumbnails( hwnd);
            return (0);

        case CAMMSG_NOTIFY:
            HandleNotification( hwnd, mp1, mp2);
            return (0);

        case CAMMSG_DLGSTATUS:
            DialogStatus( hwnd, (ULONG)mp1, (BOOL)mp2);
            return (0);

        case WM_INITMENU:
            if (mp1 == (MP)IDM_FILE) {
                STATSTUFF * pss = GetStatStuff();

                WinSendMsg( (HWND)mp2, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_SAVE, FALSE),
                    MPFROM2SHORT( MIA_DISABLED,
                        (pss->ulDelCnt || pss->ulSavCnt ? 0 : MIA_DISABLED)));
            }
            return (0);

        case WM_MENUEND:
            if (mp1 == (MP)IDM_SINGLE) {

                // pcr is actually a CAMRecPtr * but who cares?
                ULONG pcr = WinQueryWindowULong( (HWND)mp2, QWL_USER);

                if (pcr)
                    WinSendMsg( WinWindowFromID( hwnd, FID_CLIENT),
                                CM_SETRECORDEMPHASIS, (MP)pcr,
                                MPFROM2SHORT( FALSE, CRA_SOURCE));
            }
            return (0);

        case WM_SYSCOMMAND:
            if (SHORT1FROMMP( mp1) != SC_CLOSE)
                break;

            if (SaveChanges( hwnd, MAKEULONG( WM_CLOSE, 0)))
                WinPostMsg( hwnd, WM_QUIT, NULL, NULL);
            return (0);

        case WM_CHAR:
            if ((SHORT1FROMMP(mp1) & (KC_VIRTUALKEY | KC_KEYUP)) ==
                                      KC_VIRTUALKEY) {

                // if this were handled as an accelerator, you
                // couldn't delete characters when editing a title
                if (SHORT2FROMMP(mp2) == VK_DELETE) {
                    Command( hwnd, IDM_DELETE);
                    return ((MRESULT)TRUE);
                }

                if (WM_CHAR_Handler( hwnd, SHORT2FROMMP(mp2)))
                    return ((MRESULT)TRUE);
            }
            else
            if ((SHORT1FROMMP(mp1) & (KC_SCANCODE | KC_CTRL | KC_KEYUP)) ==
                                     (KC_SCANCODE | KC_CTRL) &&
                SHORT1FROMMP(mp2) && SHORT1FROMMP(mp2) < 0x100) {
                if (MatchGroupMnemonic( hwnd, SHORT1FROMMP(mp2)))
                    return ((MRESULT)TRUE);
            }
            break;

        case WM_CLOSE:
            if (SaveChanges( hwnd, MAKEULONG( WM_CLOSE, 0)))
                WinPostMsg( hwnd, WM_QUIT, NULL, NULL);
            return (0);

        case WM_DESTROY:
            TermShow( hwnd);
            break;

        case WM_INITDLG:
            InitDlg( hwnd);
            return (0);

        case WM_QUERYFRAMECTLCOUNT:
            return ((MRESULT)(SHORT1FROMMR(
                              WinDefDlgProc( hwnd, msg, mp1, mp2)) + 2));

        case WM_FORMATFRAME:
            return (FormatFrame( hwnd, mp1, mp2));

        case WM_SAVEAPPLICATION:
            if (fRestoreDefaults)
                PrfWriteProfileData( HINI_USERPROFILE, CAMINI_APP, 0, 0, 0);
            else
            {
                StoreColumns( hwnd);
                CnrStoreViews();
                StoreSort();
                WinStoreWindowPos( CAMINI_APP, CAMINI_WNDPOS, hwnd);
            }
            StoreGroups();
            break;

    } //end switch (msg)

    return (WinDefDlgProc( hwnd, msg, mp1, mp2));
}

/****************************************************************************/


void            InitDlg( HWND hwnd)

{
    BOOL    fRtn = FALSE;
    HACCEL  haccel;

do
{
    if (WinWindowFromID( hwnd, FID_CLIENT) == 0 ||
        WinWindowFromID( hwnd, IDC_STATUS) == 0 ||
        WinWindowFromID( hwnd, IDC_SIDEBAR) == 0)
        break;

    // get icons, pointers, and menu checkmarks
    InitStatStuff( hwnd);

    // set the app icon
    hAppIco = WinLoadPointer( HWND_DESKTOP, 0, IDI_APPICO);
    WinSendMsg( hwnd, WM_SETICON, (MP)hAppIco, 0);

    // init the menubar & popup menus
    if (!InitMenus( hwnd))
        break;

    // setup the container's columns, graphics, & menus
    if (CnrInit( hwnd))
        break;

    // init the sidebar windows, groups, & gbm.dll
    InitShow( hwnd);

    // init notification that the camera has been disconnected/reconnected
    if (!CAM_DEMO_FLAG)
        InitNotification( hwnd);

    // make the frame aware of the menu, status, & side bars;
    WinSendMsg( hwnd, WM_UPDATEFRAME, (MP)FCF_MENU, 0);

    // load the accelerator table for the main window
    haccel = WinLoadAccelTable( 0, 0, IDA_MAIN);
    if (!haccel || !WinSetAccelTable( 0, haccel, hwnd))
        printf( "Main - unable to load accelerator table - haccel= %x\n",
                (int)haccel);


    // wait until main() has entered its msg loop before getting any data
    WinPostMsg( hwnd, WM_COMMAND, (MP)IDM_GETDATA, 0);

    if (!WinRestoreWindowPos( CAMINI_APP, CAMINI_WNDPOS, hwnd))
        WinShowWindow( hwnd, TRUE);

    fRtn = TRUE;

} while (0);

    // if there was an error, abort
    if (!fRtn) {
        ShowInitErrorDlg( MSG_InitDisplay, IDD_ERROR);
        WinPostMsg( hwnd, WM_QUIT, 0, 0);
    }

    return;
}

/****************************************************************************/

// handles the positioning of the menu, status, & side bars

MRESULT     FormatFrame( HWND hwnd, MPARAM mp1, MPARAM mp2)

{
    USHORT  usClient = 0;
    USHORT  usCnt;
    HWND    hCnr;
    PSWP    pSWP = (PSWP)mp1;
    RECTL   rcl;

    // have the dlg format the controls it knows about
    // (which by now includes the menubar we added)
    usCnt = SHORT1FROMMR(WinDefDlgProc( hwnd, WM_FORMATFRAME, mp1, mp2));

    // find the client window's SWP
    hCnr = WinWindowFromID( hwnd, FID_CLIENT);
    while (pSWP[usClient].hwnd != hCnr)
        usClient++;

    // Status Bar

    // fill in the next SWP with some basic info
    pSWP[usCnt].hwnd = WinWindowFromID( hwnd, IDC_STATUS);
    pSWP[usCnt].fl = SWP_SIZE | SWP_MOVE | SWP_SHOW;

    // get the size of the statusbar
    WinQueryWindowRect( pSWP[usCnt].hwnd, &rcl);

    // position the statusbar at the client's current origin;
    // make it as wide as the client but leave its height as-is
    pSWP[usCnt].x  = pSWP[usClient].x;
    pSWP[usCnt].cx = pSWP[usClient].cx;
    pSWP[usCnt].y  = pSWP[usClient].y;
    pSWP[usCnt].cy = rcl.yTop;

    // reduce the client window's height & move its origin up
    // by the height of the statusbar
    pSWP[usClient].cy -= pSWP[usCnt].cy;
    pSWP[usClient].y += pSWP[usCnt].cy;

    usCnt++;

    // Side Bar

    // fill in the next SWP with some basic info
    pSWP[usCnt].hwnd = WinWindowFromID( hwnd, IDC_SIDEBAR);
    pSWP[usCnt].fl = SWP_SIZE | SWP_MOVE | SWP_SHOW;

    // get the size of the sidebar
    WinQueryWindowRect( pSWP[usCnt].hwnd, &rcl);

    // position the sidebar at the client's adjusted origin;
    // make it as tall as the client but leave its width as-is
    pSWP[usCnt].x  = pSWP[usClient].x;
    pSWP[usCnt].cx = rcl.xRight;
    pSWP[usCnt].y  = pSWP[usClient].y;
    pSWP[usCnt].cy = pSWP[usClient].cy;

    // reduce the client window's width & move its origin over
    // by the width of the sidebar
    pSWP[usClient].cx -= pSWP[usCnt].cx;
    pSWP[usClient].x += pSWP[usCnt].cx;

    usCnt++;

    // return total count of frame controls
    return( MRFROMSHORT( usCnt));
}

/****************************************************************************/

// because the sidebar buttons are children of a static control rather
// than the main dialog, keyboard navigation has to be handled "manually"

BOOL        WM_CHAR_Handler( HWND hwnd, USHORT usVK)

{
    BOOL    fRtn = FALSE;
    USHORT  usID;
    ULONG   ulCode;
    HWND    hSide;
    HWND    hFocus;
    HWND    hNext;

do {
    hSide = WinWindowFromID( hwnd, IDC_SIDEBAR);
    hFocus = WinQueryFocus( HWND_DESKTOP);
    usID = WinQueryWindowUShort( hFocus, QWS_ID);

    if (usID == FID_CLIENT) {
        if (usVK == VK_TAB)
            ulCode = EDI_FIRSTTABITEM;
        else
        if (usVK == VK_BACKTAB)
            ulCode = EDI_LASTTABITEM;
        else
            break;

        hNext = WinEnumDlgItem( hSide, hFocus, ulCode);
        WinFocusChange( HWND_DESKTOP, hNext, 0);
        fRtn = TRUE;
        break;
    }

    if (usID < IDC_GRPBTNFIRST || usID >= IDC_GRPBTNLAST)
        break;

    ulCode = 99;
    switch (usVK) {
        case VK_TAB:
            ulCode = EDI_NEXTTABITEM;
            break;

        case VK_BACKTAB:
            // if the focus is on the first button, let the
            // dialog proc move the focus to the container
            if (usID == IDC_GRPBTNFIRST)
                break;
            ulCode = EDI_PREVTABITEM;
            break;

        case VK_RIGHT:
        case VK_DOWN:
            ulCode = EDI_NEXTGROUPITEM;
            break;

        case VK_LEFT:
        case VK_UP:
            ulCode = EDI_PREVGROUPITEM;
            break;
    }

    if (ulCode == 99)
        break;

    hNext = WinEnumDlgItem( hSide, hFocus, ulCode);
    // if the focus is about to loop back to the first button,
    // let the dialog proc move the focus to the container
    if (usVK == VK_TAB && WinQueryWindowUShort( hNext, QWS_ID) == IDC_GRPBTNFIRST)
        break;
    WinFocusChange( HWND_DESKTOP, hNext, 0);
    fRtn = TRUE;

} while (0);

    return (fRtn);
}

/****************************************************************************/

// show the counters and the optional extra message

void        UpdateStatus( HWND hwnd, ULONG ulMsg)

{
    int         cnt;
    STATSTUFF * pss;
    char        szText[384];

    pss = GetStatStuff();
    cnt = sprintf( szText, pss->pszStatus,
                   pss->ulTotCnt, (pss->ulTotSize/1024),
                   pss->ulSavCnt, (pss->ulSavSize/1024),
                   pss->ulDelCnt, (pss->ulDelSize/1024));
    if (ulMsg)
        LoadString( ulMsg, &szText[cnt]);
    WinSetDlgItemText( hwnd, IDC_STATUS, (PCSZ)szText);
 
   return;
}

/****************************************************************************/
/****************************************************************************/

void        InitStatStuff( HWND hwnd)

{
    memset( &ss, 0, sizeof(STATSTUFF));
    ss.pszStatus = LoadString( STS_Status, 0);

    ss.hBlkIco = WinLoadPointer( HWND_DESKTOP, 0, IDI_BLKICO);
    ss.hSavIco = WinLoadPointer( HWND_DESKTOP, 0, IDI_SAVICO);
    ss.hDelIco = WinLoadPointer( HWND_DESKTOP, 0, IDI_DELICO);
    ss.hDoneIco = WinLoadPointer( HWND_DESKTOP, 0, IDI_DONEICO);

    return;
}

/****************************************************************************/

STATSTUFF * GetStatStuff( void)

{
    return (&ss);
}

/****************************************************************************/

void        ToggleRestoreDefaults( HWND hwnd)

{
    fRestoreDefaults ^= TRUE;
    WinSendDlgItemMsg( hwnd, FID_MENU, MM_SETITEMATTR,
                       MPFROM2SHORT( IDM_RESTOREDEF, TRUE),
                       MPFROM2SHORT( MIA_CHECKED,
                                     (fRestoreDefaults ? MIA_CHECKED : 0)));
    return;
}

/****************************************************************************/

void        DialogStatus( HWND hwnd, ULONG ulID, BOOL fOpen)

{

switch (ulID) {

    case IDD_GROUPS: {
        HWND    hTemp;

        hTemp = WinWindowFromID( hwnd, IDC_SIDEBAR);
        hTemp = WinWindowFromID( hTemp, IDC_EDITGRP);
        WinEnableWindow( hTemp, (fOpen ? FALSE : TRUE));
        break;
    }

    case IDD_SAVE: {
        fInSave = fOpen;
        if (fInSave) {
            QMSG    qmsg;

            while (WinPeekMsg( 0, &qmsg, 0, CAMMSG_SHOWTHUMB,
                               CAMMSG_SHOWTHUMB, PM_REMOVE));

            WinEnableWindow( hwnd, FALSE);
        }
        else {
            WinEnableWindow( hwnd, TRUE);
            UpdateThumbWnd( hwnd);
        }

        break;
    }

    case IDM_DETAILS:
    case IDM_SMALL:
    case IDM_MEDIUM:
    case IDM_LARGE: {
        MENUITEM    mi;

        if (!WinSendDlgItemMsg( hwnd, FID_MENU, MM_QUERYITEM,
                                MPFROM2SHORT( IDM_VIEW, FALSE),
                                (MP)&mi))
            break;

        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_COLORDER, FALSE),
                    MPFROM2SHORT( MIA_DISABLED,
                    ((ulID == IDM_DETAILS) ? 0 : MIA_DISABLED)));
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_COLRESET, FALSE),
                    MPFROM2SHORT( MIA_DISABLED,
                    ((ulID == IDM_DETAILS) ? 0 : MIA_DISABLED)));

        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_DETAILS, FALSE),
                    MPFROM2SHORT( MIA_CHECKED, 0));
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_SMALL, FALSE),
                    MPFROM2SHORT( MIA_CHECKED, 0));
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_MEDIUM, FALSE),
                    MPFROM2SHORT( MIA_CHECKED, 0));
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_LARGE, FALSE),
                    MPFROM2SHORT( MIA_CHECKED, 0));
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( ulID, FALSE),
                    MPFROM2SHORT( MIA_CHECKED, MIA_CHECKED));
        break;
    }

    case IDD_COLUMNS: {
        MENUITEM    mi;
        MPARAM      mp2;

        if (!WinSendDlgItemMsg( hwnd, FID_MENU, MM_QUERYITEM,
                                MPFROM2SHORT( IDM_VIEW, FALSE),
                                (MP)&mi))
            break;

        mp2 = MPFROM2SHORT( MIA_DISABLED, (fOpen ? MIA_DISABLED : 0));

        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_COLORDER, FALSE), mp2);
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_DETAILS, FALSE), mp2);
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_SMALL, FALSE), mp2);
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_MEDIUM, FALSE), mp2);
        WinSendMsg( mi.hwndSubMenu, MM_SETITEMATTR,
                    MPFROM2SHORT( IDM_LARGE, FALSE), mp2);
        break;
    }

    case IDD_COLORS:
        WinSendDlgItemMsg( hwnd, FID_MENU, MM_SETITEMATTR,
                           MPFROM2SHORT( IDM_COLORS, TRUE),
                           MPFROM2SHORT( MIA_DISABLED,
                                         (fOpen ? MIA_DISABLED : 0)));
        break;

} // end switch

    return;
}

/****************************************************************************/
/****************************************************************************/

// this is only used before msg loop processing begins

ULONG       ShowInitErrorDlg( ULONG ulErr, ULONG ulDlgID)

{
    char    szText[256];

    if (ulErr)
        LoadString( ulErr, szText);

    return (WinDlgBox( HWND_DESKTOP,        // parent-window
                       NULLHANDLE,          // owner-window
                       InitErrorDlgProc,    // dialog proc
                       NULLHANDLE,          // EXE module handle
                       ulDlgID,             // dialog id
                       (ulErr ? szText : 0)));  // pointer to create params
}

/****************************************************************************/

MRESULT _System InitErrorDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{
    RECTL   rcl;

    if (msg != WM_INITDLG)
        return (WinDefDlgProc( hwnd, msg, mp1, mp2));

    if (mp2)
        WinSetDlgItemText( hwnd, IDC_TEXT1, (char*)mp2);

    if (WinQueryWindowUShort( hwnd, QWS_ID) == IDD_NOCAM)
        LoadDlgStrings( hwnd, nlsNoCamDlg);
    else
        LoadDlgStrings( hwnd, nlsErrorDlg);

    WinQueryWindowRect( hwnd, &rcl);
    rcl.xLeft = WinQuerySysValue( HWND_DESKTOP, SV_CXSCREEN);
    rcl.yBottom = WinQuerySysValue( HWND_DESKTOP, SV_CYSCREEN);

    WinSetWindowPos( hwnd, 0,
                     (rcl.xLeft - rcl.xRight) / 2,
                     (rcl.yBottom - rcl.yTop) / 2,
                     0, 0, SWP_MOVE | SWP_SHOW);

    return 0;
}

/****************************************************************************/
/****************************************************************************/

HWND        OpenWaitOrExitDlg( HWND hOwner, ULONG ulDlgID)

{
    HWND    hRtn;

    hRtn = WinLoadDlg( HWND_DESKTOP,               //  parent-window
                       hOwner,                     //  owner-window
                       WaitOrExitDlgProc,          //  dialog proc
                       NULLHANDLE,                 //  EXE module handle
                       ulDlgID,                    //  dialog id
                       NULL);                      //  pointer to create params

    if (hRtn)
        WinEnableWindow( hOwner, FALSE);

    return (hRtn);
}

/****************************************************************************/

void        CloseWaitOrExitDlg( HWND hDlg)

{
    HWND    hOwner;

    if (hDlg) {
        hOwner = WinQueryWindow( hDlg, QW_OWNER);
        WinDestroyWindow( hDlg);
        WinEnableWindow( hOwner, TRUE);
    }

    return;
}

/****************************************************************************/

MRESULT _System WaitOrExitDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{
    switch (msg)
    {
        case WM_COMMAND:
            if (SHORT1FROMMP( mp1) == IDC_EXIT)
                WinPostMsg( WinQueryWindow( hwnd, QW_OWNER), WM_QUIT, 0, 0);
            return (0);

        case WM_SYSCOMMAND:
            if (SHORT1FROMMP( mp1) == SC_CLOSE) {
                WinPostMsg( WinQueryWindow( hwnd, QW_OWNER), WM_QUIT, 0, 0);
                return (0);
            }
            break;

        case WM_CLOSE:
                WinPostMsg( WinQueryWindow( hwnd, QW_OWNER), WM_QUIT, 0, 0);
            return (0);

        case WM_INITDLG: {
            HWND    hOwner;
            RECTL   rcl;
            RECTL   rclOwner;

            if (WinQueryWindowUShort( hwnd, QWS_ID) == IDD_DISCONNECT)
                LoadDlgStrings( hwnd, nlsDisconnectDlg);
            else
                LoadDlgStrings( hwnd, nlsLoadingDlg);

            hOwner = WinQueryWindow( hwnd, QW_OWNER);
            WinQueryWindowRect( hwnd, &rcl);
            WinQueryWindowRect( hOwner, &rclOwner);

            rcl.xLeft = (rclOwner.xRight - rcl.xRight) / 2;
            rcl.yBottom = (rclOwner.yTop - rcl.yTop) / 2;
            WinMapWindowPoints( hOwner, HWND_DESKTOP, (PPOINTL)&rcl, 1);
            WinSetWindowPos( hwnd, 0, rcl.xLeft, rcl.yBottom,
                             0, 0, SWP_MOVE | SWP_SHOW);

            return (0);
        }

        case WM_SAVEAPPLICATION:
            WinFocusChange( HWND_DESKTOP, WinQueryWindow( hwnd, QW_OWNER), 0);
            break;

    } //end switch (msg)

    return (WinDefDlgProc( hwnd, msg, mp1, mp2));
}

/****************************************************************************/

